# Window.py -- Pamhyr
# Copyright (C) 2023  INRAE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

import logging

from tools import trace, timer

from View.ASubWindow import ASubMainWindow
from View.ListedSubWindow import ListedSubWindow

from PyQt5.QtGui import (
    QKeySequence,
)

from PyQt5.QtCore import (
    Qt, QVariant, QAbstractTableModel,
    QCoreApplication, QModelIndex, pyqtSlot,
    QRect,
)

from PyQt5.QtWidgets import (
    QDialogButtonBox, QPushButton, QLineEdit,
    QFileDialog, QTableView, QAbstractItemView,
    QUndoStack, QShortcut, QAction, QItemDelegate,
    QComboBox, QVBoxLayout, QHeaderView, QTabWidget,
)

from View.LateralContribution.UndoCommand import (
    SetNameCommand, SetEdgeCommand, SetTypeCommand,
    AddCommand, DelCommand, SortCommand,
    MoveCommand, PasteCommand, DuplicateCommand,
)

from Model.LateralContribution.LateralContributionTypes import (
    NotDefined, LateralContrib, Rain, Evaporation,
)

from View.LateralContribution.Table import (
    TableModel, ComboBoxDelegate
)

from View.Plot.MplCanvas import MplCanvas
from View.Geometry.PlotXY import PlotXY
from View.LateralContribution.translate import *
from View.LateralContribution.Edit.Window import EditLateralContributionWindow

_translate = QCoreApplication.translate

logger = logging.getLogger()

class LateralContributionWindow(ASubMainWindow, ListedSubWindow):
    def __init__(self, title="Lateral contribution", study=None, parent=None):
        self._title = title + " - " + study.name

        super(LateralContributionWindow, self).__init__(
            name=title, ui="LateralContributions", parent=parent
        )

        self._study = study
        self._lcs = self._study.river.lateral_contribution

        self.setup_sc()
        self.setup_table()
        self.setup_graph()
        self.setup_connections()

        self.ui.setWindowTitle(self._title)

    def setup_sc(self):
        self._undo_stack = QUndoStack()

        self.undo_sc = QShortcut(QKeySequence.Undo, self)
        self.redo_sc = QShortcut(QKeySequence.Redo, self)
        self.copy_sc = QShortcut(QKeySequence.Copy, self)
        self.paste_sc = QShortcut(QKeySequence.Paste, self)

    def setup_table(self):
        self._table = {}

        for t in ["liquid", "solid", "suspenssion"]:
            table = self.find(QTableView, f"tableView_{t}")
            self._table[t] = TableModel(
                data = self._study.river,
                undo = self._undo_stack,
                tab = t,
            )
            table.setModel(self._table[t])

            self._delegate_type = ComboBoxDelegate(
                data = self._study.river,
                mode = "type",
                tab = t,
                parent=self
            )
            self._delegate_edge = ComboBoxDelegate(
                data = self._study.river,
                mode = "edge",
                tab = t,
                parent=self
            )

            table.setItemDelegateForColumn(
                1, self._delegate_type
            )
            table.setItemDelegateForColumn(
                2, self._delegate_edge
            )

            table.setSelectionBehavior(QAbstractItemView.SelectRows)
            table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
            table.setAlternatingRowColors(True)

    def setup_graph(self):
        self.canvas = MplCanvas(width=5, height=4, dpi=100)
        self.canvas.setObjectName("canvas")
        self.plot_layout = self.find(QVBoxLayout, "verticalLayout")
        self.plot_layout.addWidget(self.canvas)

        self.plot = PlotXY(
            canvas = self.canvas,
            data = None,
            toolbar = None,
        )

    def setup_connections(self):
        self.find(QAction, "action_add").triggered.connect(self.add)
        self.find(QAction, "action_del").triggered.connect(self.delete)
        self.find(QAction, "action_edit").triggered.connect(self.edit)
        self.find(QAction, "action_sort").triggered.connect(self.sort)

        self.undo_sc.activated.connect(self.undo)
        self.redo_sc.activated.connect(self.redo)
        self.copy_sc.activated.connect(self.copy)
        self.paste_sc.activated.connect(self.paste)

        for t in ["liquid", "solid", "suspenssion"]:
            table = self.find(QTableView, f"tableView_{t}")
            table.selectionModel()\
                 .selectionChanged\
                 .connect(self._set_current_reach)

            self._table[t].dataChanged\
                          .connect(self._set_current_reach)

    def current_tab(self):
        return self.find(QTabWidget, "tabWidget")\
                   .currentWidget()\
                   .objectName()\
                   .replace("tab_", "")

    def index_selected_row(self):
        tab = self.current_tab()
        table = self.find(QTableView, f"tableView_{tab}")
        return table.selectionModel()\
                    .selectedRows()[0]\
                    .row()

    def index_selected_rows(self):
        tab = self.current_tab()
        table = self.find(QTableView, f"tableView_{tab}")
        return list(
            # Delete duplicate
            set(
                map(
                    lambda i: i.row(),
                    table.selectedIndexes()
                )
            )
        )

    def _set_current_reach(self):
        tab = self.current_tab()
        rows = self.index_selected_rows()

        data = None
        highlight = None

        if len(rows) > 0:
            edge = self._study\
                       .river\
                       .lateral_contribution\
                       .get(tab, rows[0])\
                       .edge
            if edge:
                data = edge.reach
                lc = self._lcs.get(tab, rows[0])
                highlight = (lc.begin_kp, lc.end_kp)

        self.plot = PlotXY(
            canvas = self.canvas,
            data = data,
            toolbar = None,
        )
        self.plot.draw(highlight=highlight)

    def add(self):
        tab = self.current_tab()
        rows = self.index_selected_rows()
        if self._lcs.len(tab) == 0 or len(rows) == 0:
            self._table[tab].add(0)
        else:
            self._table[tab].add(rows[0])

        self._set_current_reach()

    def delete(self):
        tab = self.current_tab()
        rows = self.index_selected_rows()
        if len(rows) == 0:
            return

        self._table[tab].delete(rows)
        self._set_current_reach()

    def sort(self):
        tab = self.current_tab()
        self._table[tab].sort(False)
        self._set_current_reach()

    def move_up(self):
        tab = self.current_tab()
        row = self.index_selected_row()
        self._table[tab].move_up(row)
        self._set_current_reach()

    def move_down(self):
        tab = self.current_tab()
        row = self.index_selected_row()
        self._table[tab].move_down(row)
        self._set_current_reach()

    def copy(self):
        logger.info("TODO: copy")
        self._set_current_reach()

    def paste(self):
        logger.info("TODO: paste")
        self._set_current_reach()

    def undo(self):
        tab = self.current_tab()
        self._table[tab].undo()
        self._set_current_reach()

    def redo(self):
        tab = self.current_tab()
        self._table[tab].redo()
        self._set_current_reach()

    def edit(self):
        tab = self.current_tab()
        rows = self.index_selected_rows()
        for row in rows:
            win = self.sub_win_filter_first(
                "Edit lateral contribution",
                contain = [f"({self._lcs.get(tab, row).id})"]
            )

            if win is None:
                win = EditLateralContributionWindow(
                    data=self._lcs.get(tab, row),
                    study=self._study,
                    parent=self
                )
                win.show()
            else:
                win.activateWindow()
