# -*- coding: utf-8 -*-

import logging

from tools import trace, timer

from PyQt5.QtCore import (
    Qt, QVariant, QAbstractTableModel,
    QCoreApplication, QModelIndex, pyqtSlot,
    QRect,
)

from PyQt5.QtWidgets import (
    QDialogButtonBox, QPushButton, QLineEdit,
    QFileDialog, QTableView, QAbstractItemView,
    QUndoStack, QShortcut, QAction, QItemDelegate,
    QComboBox,
)

from View.SedimentLayers.Reach.UndoCommand import *
from View.SedimentLayers.Reach.translate import *

_translate = QCoreApplication.translate

logger = logging.getLogger()

class ComboBoxDelegate(QItemDelegate):
    def __init__(self, study=None, parent=None):
        super(ComboBoxDelegate, self).__init__(parent)

        self._study = study

    def createEditor(self, parent, option, index):
        self.editor = QComboBox(parent)

        self.editor.addItems(
            [_translate("SedimentLayers", "Not defined")] +
            list(
                map(
                    lambda sl: str(sl),
                    self._study.river.sediment_layers.sediment_layers
                )
            )
        )

        self.editor.setCurrentText(index.data(Qt.DisplayRole))
        return self.editor

    def setEditorData(self, editor, index):
        value = index.data(Qt.DisplayRole)
        self.editor.currentTextChanged.connect(self.currentItemChanged)

    def setModelData(self, editor, model, index):
        text = str(editor.currentText())
        model.setData(index, text)
        editor.close()
        editor.deleteLater()

    def updateEditorGeometry(self, editor, option, index):
        r = QRect(option.rect)
        if self.editor.windowFlags() & Qt.Popup and editor.parent() is not None:
            r.setTopLeft(self.editor.parent().mapToGlobal(r.topLeft()))
        editor.setGeometry(r)

    @pyqtSlot()
    def currentItemChanged(self):
        self.commitData.emit(self.sender())


class TableModel(QAbstractTableModel):
    def __init__(self, study=None, reach=None, undo=None):
        super(QAbstractTableModel, self).__init__()
        self._headers = list(table_headers.keys())
        self._study = study
        self._undo = undo
        self._reach = reach

    def flags(self, index):
        column = index.column()

        options = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if self._headers[column] == "sl":
            options |= Qt.ItemIsEditable

        return options

    def rowCount(self, parent):
        return self._reach.number_profiles

    def columnCount(self, parent):
        return len(self._headers)

    def data(self, index, role):
        if role != Qt.ItemDataRole.DisplayRole:
            return QVariant()

        row = index.row()
        column = index.column()

        if self._headers[column] == "name":
            return self._reach.profile(row).name
        if self._headers[column] == "kp":
            return self._reach.profile(row).kp
        if self._headers[column] == "sl":
            value = self._reach.profile(row).sl
            if value == None:
                text = _translate("SedimentLayers", "Not defined")
                return text
            return str(value)

        return QVariant()

    def headerData(self, friction, orientation, role):
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return table_headers[self._headers[friction]]

        return QVariant()

    def setData(self, index, value, role=Qt.EditRole):
        if not index.isValid() or role != Qt.EditRole:
            return False

        row = index.row()
        column = index.column()

        if self._headers[column] == "sl":
            new = None
            if value != _translate("SedimentLayers", "Not defined"):
                new = next(
                    filter(
                        lambda sl: str(sl) == value,
                        self._study.river.sediment_layers.sediment_layers
                    )
                )

            self._undo.push(
                SetSLCommand(
                    self._reach, row, new
                )
            )

        self.dataChanged.emit(index, index)
        return True

    def apply_sl_each_profile(self, sl):
        self._undo.push(
            ApplySLCommand(
                self._reach, sl
            )
        )
        self.layoutChanged.emit()

    def undo(self):
        self._undo.undo()
        self.layoutChanged.emit()

    def redo(self):
        self._undo.redo()
        self.layoutChanged.emit()
