# -*- coding: utf-8 -*-

import logging

from tools import trace, timer

from View.ASubWindow import ASubMainWindow
from View.ListedSubWindow import ListedSubWindow

from PyQt5.QtGui import (
    QKeySequence,
)

from PyQt5.QtCore import (
    Qt, QVariant, QAbstractTableModel,
    QCoreApplication, QModelIndex, pyqtSlot,
    QRect,
)

from PyQt5.QtWidgets import (
    QDialogButtonBox, QPushButton, QLineEdit,
    QFileDialog, QTableView, QAbstractItemView,
    QUndoStack, QShortcut, QAction, QItemDelegate,
    QComboBox, QVBoxLayout, QHeaderView, QTabWidget,
)

from View.SedimentLayers.Reach.UndoCommand import *
from View.SedimentLayers.Reach.Table import *
from View.SedimentLayers.Reach.Plot import Plot
from View.SedimentLayers.Reach.SLDialog import SLDialog

from View.Plot.MplCanvas import MplCanvas
from View.Plot.navigation_toolbar_2qt import PamHyrNavigationToolbar2QT

from View.SedimentLayers.Reach.translate import *
from View.SedimentLayers.Window import SedimentLayersWindow
from View.SedimentLayers.Reach.Profile.Window import ProfileSedimentLayersWindow

_translate = QCoreApplication.translate

logger = logging.getLogger()

class ReachSedimentLayersWindow(ASubMainWindow, ListedSubWindow):
    def __init__(self, title="Reach sediment layers", study=None, parent=None):
        self._study = study
        self._sediment_layers = self._study.river.sediment_layers
        self._reach = self._study.river.current_reach().reach

        self.setup_title(title)

        super(ReachSedimentLayersWindow, self).__init__(
            name=self._title, ui="ReachSedimentLayers", parent=parent
        )

        self.setup_sc()
        self.setup_table()
        self.setup_graph()
        self.setup_connections()

        self.ui.setWindowTitle(self._title)

    def setup_title(self, title):
        self._title = (
            title + " - " + self._study.name + " - " + self._reach.name
        )

    def setup_sc(self):
        self._undo_stack = QUndoStack()

        self.undo_sc = QShortcut(QKeySequence.Undo, self)
        self.redo_sc = QShortcut(QKeySequence.Redo, self)
        self.copy_sc = QShortcut(QKeySequence.Copy, self)
        self.paste_sc = QShortcut(QKeySequence.Paste, self)

    def setup_table(self):
        table = self.find(QTableView, f"tableView")
        self._table = TableModel(
            study = self._study,
            reach = self._reach,
            undo = self._undo_stack,
        )
        table.setModel(self._table)

        self._delegate_stricklers = ComboBoxDelegate(
            study = self._study,
            parent=self
        )

        table.setItemDelegateForColumn(
            list(table_headers).index("sl"),
            self._delegate_stricklers
        )

        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        table.setAlternatingRowColors(True)

    def setup_graph(self):
        self.canvas = MplCanvas(width=5, height=4, dpi=100)
        self.canvas.setObjectName("canvas")
        self.toolbar = PamHyrNavigationToolbar2QT(
            self.canvas, self
        )
        self.plot_layout = self.find(QVBoxLayout, "verticalLayout_2")
        self.plot_layout.addWidget(self.toolbar)
        self.plot_layout.addWidget(self.canvas)

        self._update_plot()

    def _update_plot(self):
        self.plot = Plot(
            canvas = self.canvas,
            data = self._reach,
            toolbar = self.toolbar,
            display_current = False
        )
        self.plot.draw()


    def setup_connections(self):
        self.find(QAction, "action_edit").triggered.connect(self.edit_profile)

        self.find(QPushButton, "pushButton_edit")\
            .clicked\
            .connect(self.edit_sl)
        self.find(QPushButton, "pushButton_apply")\
            .clicked\
            .connect(self.apply_sl_each_profile)

        self.undo_sc.activated.connect(self.undo)
        self.redo_sc.activated.connect(self.redo)
        self.copy_sc.activated.connect(self.copy)
        self.paste_sc.activated.connect(self.paste)

        self._table.layoutChanged\
                   .connect(self._update_plot)
        self._table.dataChanged\
                   .connect(self._update_plot)


    def index_selected_rows(self):
        table = self.find(QTableView, f"tableView")
        return list(
            # Delete duplicate
            set(
                map(
                    lambda i: i.row(),
                    table.selectedIndexes()
                )
            )
        )

    def copy(self):
        logger.info("TODO: copy")

    def paste(self):
        logger.info("TODO: paste")

    def undo(self):
        self._table.undo()
        self._update_plot()

    def redo(self):
        self._table.redo()
        self._update_plot()

    def apply_sl_each_profile(self):
        slw = SLDialog(
            study = self._study,
            parent = self
        )
        if slw.exec():
            sl = slw.sl
            self._table.apply_sl_each_profile(sl)

        self._update_plot()

    def edit_profile(self):
        rows = self.index_selected_rows()

        for row in rows:
            slw = ProfileSedimentLayersWindow(
                study = self._study,
                profile = self._reach.profile(row),
                parent = self
            )
            slw.show()


    def edit_sl(self):
        slw = SedimentLayersWindow(
            study = self._study,
            parent = self
        )
        slw.show()
