# Window.py -- Pamhyr
# Copyright (C) 2023  INRAE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

import logging

from tools import trace, timer

from View.ASubWindow import ASubMainWindow
from View.ListedSubWindow import ListedSubWindow

from PyQt5.QtGui import (
    QKeySequence,
)

from PyQt5.QtCore import (
    Qt, QVariant, QAbstractTableModel,
    QCoreApplication, QModelIndex, pyqtSlot,
    QRect,
)

from PyQt5.QtWidgets import (
    QDialogButtonBox, QPushButton, QLineEdit,
    QFileDialog, QTableView, QAbstractItemView,
    QUndoStack, QShortcut, QAction, QItemDelegate,
    QComboBox, QVBoxLayout, QHeaderView, QTabWidget,
)

from Model.Stricklers.Stricklers import Stricklers

from View.Stricklers.UndoCommand import PasteCommand
from View.Stricklers.Table import TableModel

_translate = QCoreApplication.translate

logger = logging.getLogger()

class StricklersWindow(ASubMainWindow, ListedSubWindow):
    def __init__(self, title="Stricklers", study=None, config=None, parent=None):
        self._title = title + " - " + study.name

        super(StricklersWindow, self).__init__(
            name=title, ui="Stricklers", parent=parent
        )

        self._study = study
        self._config = config

        self.setup_sc()
        self.setup_table()
        self.setup_connections()

        self.ui.setWindowTitle(self._title)

    def setup_sc(self):
        self._undo_stack = QUndoStack()

        self.undo_sc = QShortcut(QKeySequence.Undo, self)
        self.redo_sc = QShortcut(QKeySequence.Redo, self)
        self.copy_sc = QShortcut(QKeySequence.Copy, self)
        self.paste_sc = QShortcut(QKeySequence.Paste, self)

    def setup_table(self):
        self._table = {}

        for t in ["app", "study"]:
            table = self.find(QTableView, f"tableView_{t}")
            if t == "study":
                data = self._study.river.stricklers
            else:
                data = self._config.stricklers

            self._table[t] = TableModel(
                data = data,
                undo = self._undo_stack,
            )

            table.setModel(self._table[t])
            table.setSelectionBehavior(QAbstractItemView.SelectRows)
            table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
            table.setAlternatingRowColors(True)

    def setup_connections(self):
        self.find(QAction, "action_add").triggered.connect(self.add)
        self.find(QAction, "action_del").triggered.connect(self.delete)
        self.find(QAction, "action_sort").triggered.connect(self.sort)

        self.undo_sc.activated.connect(self.undo)
        self.redo_sc.activated.connect(self.redo)
        self.copy_sc.activated.connect(self.copy)
        self.paste_sc.activated.connect(self.paste)

    def index_selected_rows(self):
        table = self.find(QTableView, f"tableView_study")
        return list(
            set(
                map(
                    lambda i: i.row(),
                    table.selectedIndexes()
                )
            )
        )


    def add(self):
        rows = self.index_selected_rows()
        if len(rows) == 0:
            self._table['study'].add(0)
        else:
            self._table['study'].add(rows[0])

    def delete(self):
        rows = self.index_selected_rows()
        if len(rows) == 0:
            return

        self._table['study'].delete(rows)

    def sort(self):
        self._table['study'].sort(False)

    def copy(self):
        logger.info("TODO: copy")

    def paste(self):
        logger.info("TODO: paste")

    def undo(self):
        self._table['study'].undo()

    def redo(self):
        self._table['study'].redo()
